"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.makeAccessor = makeAccessor;
exports.makeAccessor2 = makeAccessor2;
exports.getValue = getValue;
exports.datasetsFromPropsOrDescendants = datasetsFromPropsOrDescendants;
exports.inferDataType = inferDataType;
exports.inferDatasetsType = inferDatasetsType;
exports.isValidDomain = isValidDomain;
exports.combineDomains = combineDomains;
exports.combineBorderObjects = combineBorderObjects;
exports.domainFromData = domainFromData;
exports.getDataDomainByAxis = getDataDomainByAxis;
exports.domainFromDatasets = domainFromDatasets;
exports.domainFromRangeData = domainFromRangeData;
exports.combineDatasets = combineDatasets;

var _isFunction = _interopRequireDefault(require("lodash/isFunction"));

var _isNull = _interopRequireDefault(require("lodash/isNull"));

var _isUndefined = _interopRequireDefault(require("lodash/isUndefined"));

var _property = _interopRequireDefault(require("lodash/property"));

var _isNumber = _interopRequireDefault(require("lodash/isNumber"));

var _isDate = _interopRequireDefault(require("lodash/isDate"));

var _uniq = _interopRequireDefault(require("lodash/uniq"));

var _flatten = _interopRequireDefault(require("lodash/flatten"));

var _compact = _interopRequireDefault(require("lodash/compact"));

var _fromPairs = _interopRequireDefault(require("lodash/fromPairs"));

var _get = _interopRequireDefault(require("lodash/get"));

var _maxBy = _interopRequireDefault(require("lodash/maxBy"));

var _keyBy = _interopRequireDefault(require("lodash/keyBy"));

var _uniqBy = _interopRequireDefault(require("lodash/uniqBy"));

var _has = _interopRequireDefault(require("lodash/has"));

var _forEach = _interopRequireDefault(require("lodash/forEach"));

var _identity = _interopRequireDefault(require("lodash/identity"));

var _d3Array = require("d3-array");

var _react = _interopRequireDefault(require("react"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * `makeAccessor` creates an accessor or "getter" function given a variety of options
 * to be used for retrieving a data value from within an object or array
 *
 * If given a function, it is passed through.
 * If given null or undefined, the getter is the identity function - ie. returns whatever it's passed
 * If given an array index or deep object key string, the value will be retrieved using _.property
 *
 * @example
 * makeAccessor(null)(4); // 4
 * makeAccessor(d => d + 1)(4); // 5
 * makeAccessor(1)(['a', 'b', 'c']); // 'b'
 * makeAccessor('x.0.y')({x: [{y: 9}]}); // 9
 *
 * @param {any} key - Getter, which may be a function, integer, string, null, or undefined;
 * @returns {function} accessor - Accessor function
 */
function makeAccessor(key) {
  return (0, _isFunction.default)(key) ? key : (0, _isNull.default)(key) || (0, _isUndefined.default)(key) ? _identity.default : (0, _property.default)(key);
}
/**
 * `makeAccessor2` creates a constant accessor function if passed a value,
 * if passed a function, just returns it
 */


function makeAccessor2(valueOrAccessor) {
  if ((0, _isFunction.default)(valueOrAccessor)) return valueOrAccessor;
  return () => valueOrAccessor;
}
/**
 * `getValue` takes as its first argument a value or an accessor function.
 * If it's a value (ie not a function), it is returned.
 * If a function, returns the result of calling function with remaining arguments
 */


function getValue(accessor, ...args) {
  return (0, _isFunction.default)(accessor) ? accessor(...args) : accessor;
}
/**
 * `datasetsFromPropsOrDescendants` expects a `props` object which may have `children`.
 * if `props` has `data` or `datasets`, returns it; otherwise recursively searches props.children
 * for components have `data` or `datasets` and combines them into one `datasets` array.
 * `props.data` is wrapped in array because this returns `datasets` (multiple arrays of `data`)
 *
 * @param {Object} props - A React props object, which may have `children` with their own props.
 * @returns {Array.<Array>} datasets - An array of arrays of data objects
 */


function datasetsFromPropsOrDescendants(props) {
  if (Array.isArray(props.datasets)) {
    return props.datasets;
  } else if (Array.isArray(props.data)) {
    return [props.data];
  } else if (_react.default.Children.count(props.children)) {
    let datasets = []; // use Children.forEach instead of map, because Children.map flattens the arrays

    _react.default.Children.forEach(props.children, child => {
      datasets = datasets.concat(datasetsFromPropsOrDescendants(child.props));
    });

    return datasets;
  }

  return [];
}

function inferDataType(data, accessor = _identity.default) {
  if (!Array.isArray(data)) throw new Error('inferDataType expects a data array');else if (data.every((d, i) => accessor(d, i) === undefined)) return 'categorical'; // should this be allowed?
  else if (data.every((d, i) => (0, _isNumber.default)(accessor(d, i)) || accessor(d, i) === undefined)) return 'number';else if (data.every((d, i) => (0, _isDate.default)(accessor(d, i)) || accessor(d, i) === undefined)) return 'time';else return 'categorical';
}

function inferDatasetsType(datasets, accessor = _identity.default) {
  if (!Array.isArray(datasets)) throw new Error('inferDatasetsType expects a datasets array');
  const types = datasets.map(data => inferDataType(data, accessor));
  const uniqTypes = (0, _uniq.default)(types);
  return uniqTypes.length === 1 ? uniqTypes[0] : 'categorical';
}

function isValidDomain(domain, type = 'categorical') {
  return Array.isArray(domain) && !!domain.length && ( // categorical domain can be any array of anything
  type === 'categorical' || // number/time domains should look like [min, max]
  type === 'number' && domain.length === 2 && domain.every(_isNumber.default) || type === 'time' && domain.length === 2 && domain.every(_isDate.default));
}

function combineDomains(domains, dataType) {
  if (!Array.isArray(domains)) return undefined;
  return dataType === 'categorical' ? (0, _uniq.default)((0, _flatten.default)((0, _compact.default)(domains))) : (0, _d3Array.extent)((0, _flatten.default)(domains));
}

function combineBorderObjects(borderObjects) {
  return (0, _fromPairs.default)(['top', 'bottom', 'left', 'right'].map(k => {
    // combine border objects by taking the max value of each spacing direction
    return [k, (0, _get.default)((0, _maxBy.default)(borderObjects, k), k)];
  }));
}

function domainFromData(data, accessor = _identity.default, type = undefined) {
  let typeToUse = type;
  if (!typeToUse) typeToUse = inferDataType(data, accessor);
  return typeToUse === 'number' || typeToUse === 'time' ? (0, _d3Array.extent)(data.map(accessor)) : (0, _uniq.default)(data.map(accessor));
}

function getDataDomainByAxis(props) {
  const {
    horizontal,
    data,
    getX,
    getY
  } = props;
  const accessor = horizontal ? makeAccessor(getY) : makeAccessor(getX);
  const rangeAxis = horizontal ? 'y' : 'x';
  return {
    [rangeAxis]: domainFromData(data, accessor)
  };
}

function domainFromDatasets(datasets, accessor = _identity.default, type = undefined) {
  // returns the default domain of a collection of datasets with an accessor function
  // for numeric and date-type datasets, returns the extent (min and max) of the numbers/dates
  // for categorical datasets, returns the set of distinct category values
  let typeToUse = type;
  if (!typeToUse) typeToUse = inferDatasetsType(datasets, accessor);
  const domains = datasets.map(data => domainFromData(data, accessor, typeToUse));
  return combineDomains(domains, typeToUse);
}

function domainFromRangeData(data, rangeStartAccessor, rangeEndAccessor, dataType) {
  // returns the domain of dataset for which each datum represents a range of values
  // ie. has a start and end value rather than a single value
  // for example, time ranges
  let dataTypeToUse = dataType;
  if (!dataTypeToUse) dataTypeToUse = inferDataType(data, rangeStartAccessor);

  switch (dataTypeToUse) {
    case 'number':
    case 'time':
      return (0, _d3Array.extent)((0, _flatten.default)([(0, _d3Array.extent)(data, (d, i) => +rangeStartAccessor(d, i)), (0, _d3Array.extent)(data, (d, i) => +rangeEndAccessor(d, i))]));

    case 'categorical':
      return (0, _uniq.default)((0, _flatten.default)([data.map(rangeStartAccessor), data.map(rangeEndAccessor)]));

    default:
      return [];
  }
}

function combineDatasets(datasetsInfo = [], combineKey = 'x') {
  // combineDatasets combines multiple datasets into one, joined on a common key 'combineKey'
  // datasetsInfo is an array that looks like:
  // [
  //   {data: [{x: 0, y: 3}, ...], combineKey: 'x', dataKeys: {y: 'y0'}}
  //   {data: [{count: 0, value: 4}], combineKey: 'count', dataKeys: {value: 'y1'}}
  // ]
  // where `data` is an array of data points of any shape
  // `combineKey` is the key for the value which the datasets are joined on
  // `dataKeys` are getters for other values in each datapoint which should be merged into the combined dataset
  //   - key = getter in original datapoint, value = setter for combined dataset
  // example above (with default combinedKey) results in:
  // [{x: 0, y0: 3, y1: 4}, ...]
  // index each dataset by its combineKey values so we can quickly lookup if it has data for a given value
  const datasetLookups = datasetsInfo.map(datasetInfo => {
    const {
      data
    } = datasetInfo;
    return (0, _keyBy.default)(data, datasetInfo.combineKey || combineKey);
  }); // Grab combineKey from each dataset and flatten into one array

  const allCombineValues = [].concat(...datasetsInfo.map(datasetInfo => datasetInfo.data.map(makeAccessor(datasetInfo.combineKey || combineKey)))); // Get all unique values

  const uniqueValues = (0, _uniqBy.default)(allCombineValues, value => {
    return value instanceof Date ? value.toString() : value;
  }); // For each of the unique combineKey data values, go through each dataset and look for a combineKey value that matches
  // if we find it, combine the values for that datum's dataKeys into the final combinedDatum object

  return uniqueValues.map(combineValue => {
    const combinedDatum = {
      [combineKey]: combineValue
    };
    datasetsInfo.forEach((datasetInfo, datasetIndex) => {
      if (!datasetInfo.dataKeys || !Object.keys(datasetInfo.dataKeys).length) return;
      const datasetLookup = datasetLookups[datasetIndex];
      if (!(0, _has.default)(datasetLookup, combineValue)) return;
      const datum = datasetLookup[combineValue];
      (0, _forEach.default)(datasetInfo.dataKeys, (newDataKey, originalDataKey) => {
        combinedDatum[newDataKey] = datum[originalDataKey];
      });
    });
    return combinedDatum;
  });
}
//# sourceMappingURL=Data.js.map